# Test the formula of Trench, Krattenthaler, and Elouafi,
# with random rational numbers.

def random_rational_list(n):
   s = set()
   while len(s) < n:
      q = QQ.random_element()
      s.add(q)
   return list(s)

def random_rational_list_special(n):
   s = set()
   while len(s) < n:
      q = QQ.random_element()
      if (q not in [0, 1, -1]) and (q not in s) and (1 / q not in s):
         s.add(q)
   return list(s)

def random_rational_list_very_special(n):
   s = set()
   while len(s) < n:
      q = QQ.random_element()
      test0 = q not in [0, 1, -1]
      test1 = (q not in s) and (- q not in s)
      if test0 and test1:
         s.add(q)
   return list(s)

def prod_dif(x, j):
   return prod(x[j] - x[k] for k in range(j) + range(j + 1, len(x)))

def hom_via_geom_progr(x, m):
   n = len(x)
   return sum(x[j] ** (m + n - 1) / prod_dif(x, j) for j in range(n))

def matrix_schur_jt1(x, la):
   n = len(x)
   F = x[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         M[j, k] = hom_via_geom_progr(x, r) if r >= 0 else 0
   return M

def schur_via_jt1(x, la):
   return det(matrix_schur_jt1(x, la))

def matrix_symplectic_schur_jt1(x, la):
   n = len(x)
   F = x[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k
         first_summand = hom_via_geom_progr(x, r) if r >= 0 else 0
         second_summand = hom_via_geom_progr(x, s) if s >= 0 else 0
         a = first_summand + second_summand
         M[j, k] = a if k > 0 else a / 2
   return M

def symplectic_schur_via_jt1(x, la):
   return det(matrix_symplectic_schur_jt1(x, la))

def matrix_orthogonal_schur_jt1(x, la):
   n = len(x)
   F = x[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k - 2
         first_summand = hom_via_geom_progr(x, r) if r >= 0 else 0
         second_summand = hom_via_geom_progr(x, s) if s >= 0 else 0
         M[j, k] = first_summand - second_summand
   return M

def orthogonal_schur_via_jt1(x, la):
   return det(matrix_orthogonal_schur_jt1(x, la))

def test_krattenthaler_numer(n, m, verbos):
   # n = semiwidth, m = size of the matrix
   if verbos >= 2:
      print('test_krattenthaler_numer, n = %d, m = %d' % (n, m))
   x = random_rational_list_special(n)
   y = list(x) + [1 / x[j] for j in range(n)]
   yplus = y + [1]
   yminus = y + [-1]
   F = QQ
   y = [F(yelem) for yelem in y]
   yplus = [F(yelem) for yelem in yplus]
   yminus = [F(yelem) for yelem in yminus]
   lhs = schur_via_jt1(y, [m] * n)
   p = ceil(m / 2)
   if mod(m, 2) == 0:
      factor1 = orthogonal_schur_via_jt1(yplus, [p] * n)
      factor2 = orthogonal_schur_via_jt1(yminus, [p] * n)
   else:
      factor1 = symplectic_schur_via_jt1(y, [p - 1] * n)
      factor2 = orthogonal_schur_via_jt1(y, [p] * n)
   rhs = factor1 * factor2
   if verbos >= 3:
      print('lhs = ' + str(lhs))
      print('rhs = ' + str(rhs))
   return lhs == rhs

def big_test_krattenthaler_numer(nmax, mmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_krattenthaler_numer, nmax = %d, mmax = %d, nrep = %d' % (nmax, mmax, nrep))
   tasks = [(n, m, r) for n in range(1, nmax + 1) for m in range(mmax + 1) for r in range(nrep)]
   return all(test_krattenthaler_numer(n, m, verbos) for (n, m, r) in tasks)

# test Elouafi's formula,
# computing sz, spz, and oz via Jacobi--Trudi formulas

def sum_chebyshev_U(m, t):
   return sum([chebyshev_U(k, t) for k in range(m + 1)])

def homz(z, m):
   n = len(z)
   return sum([chebyshev_U(n + m - 1, z[j] / 2) / prod_dif(z, j) for j in range(n)])

def homz_odd(z, m):
   n = len(z)
   return sum([sum_chebyshev_U(n + m - 1, z[j] / 2) / prod_dif(z, j) for j in range(n)])

def sz_via_jt1(z, la):
   F = z[0].parent()
   n = len(z)
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         M[j, k] = homz(z, r) if r >= 0 else 0
   return det(M)

def spz_via_jt1(z, la):
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k
         first_summand = homz(z, r) if r >= 0 else 0
         second_summand = homz(z, s) if s >= 0 else 0
         a = first_summand + second_summand
         M[j, k] = a if k > 0 else a / 2
   return det(M)

def oz_via_jt1(z, la):
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k - 2
         first_summand = homz(z, r) if r >= 0 else 0
         second_summand = homz(z, s) if s >= 0 else 0
         M[j, k] = first_summand - second_summand
   return det(M)

def oz_odd_via_jt1(z, la):
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k - 2
         first_summand = homz_odd(z, r) if r >= 0 else 0
         second_summand = homz_odd(z, s) if s >= 0 else 0
         M[j, k] = first_summand - second_summand
   return det(M)

def oz_odd_neg_via_jt1(z, la):
   zneg = [-z[j] for j in range(len(z))]
   sign_la = (-1) ** sum(la)
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k - 2
         first_summand = homz_odd(zneg, r) if r >= 0 else 0
         second_summand = homz_odd(zneg, s) if s >= 0 else 0
         M[j, k] = first_summand - second_summand
   return sign_la * det(M)

def spz_odd_via_jt1(z, la):
   F = z[0].parent()
   p = len(la)
   M = matrix(F, p, p)
   for j in range(p):
      for k in range(p):
         r = la[j] - j + k
         s = la[j] - j - k
         first_summand = homz_odd(z, r) if r >= 0 else 0
         second_summand = homz_odd(z, s) if s >= 0 else 0
         a = first_summand + second_summand
         M[j, k] = a if k > 0 else a / 2
   return det(M)

def test_elouafi_numer(n, m, verbos):
   # n = semiwidth, m = size of the matrix
   if verbos >= 2:
      print('test_elouafi_numer, n = %d, m = %d' % (n, m))
   z = random_rational_list_special(n)
   lhs = sz_via_jt1(z, [m] * n)
   p = ceil(m / 2)
   if mod(m, 2) == 0:
      factor1 = oz_odd_via_jt1(z, [p] * n)
      factor2 = oz_odd_neg_via_jt1(z, [p] * n)
   else:
      factor1 = spz_via_jt1(z, [p - 1] * n)
      factor2 = oz_via_jt1(z, [p] * n)
   rhs = factor1 * factor2
   if verbos >= 3:
      print('lhs = ' + str(lhs))
      print('rhs = ' + str(rhs))
   return lhs == rhs

def big_test_elouafi_numer(nmax, mmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_elouafi_numer, nmax = %d, mmax = %d' % (nmax, mmax))
   tasks = [(n, m, r) for n in range(1, nmax + 1) for m in range(mmax + 1) for r in range(nrep)]
   return all(test_elouafi_numer(n, m, verbos) for (n, m, r) in tasks)

# test Trench's formula

def udet(x, powers):
   n = len(x)
   F = x[0].parent()
   M = matrix(F, n, n)
   for j in range(n):
      for k in range(n):
         M[j, k] = chebyshev_U(powers[j], x[k] / 2)
   return det(M)

def tdet(x, powers):
   n = len(x)
   F = x[0].parent()
   M = matrix(F, n, n)
   for j in range(n):
      for k in range(n):
         M[j, k] = 2 * chebyshev_T(powers[j], x[k] / 2) if powers[j] > 0 else 1
   return det(M)

def test_trench_numer(n, m, verbos):
   # n = semiwidth, m = size of the matrix
   if verbos >= 2:
      print('test_trench_numer, n = %d, m = %d' % (n, m))
   u = random_rational_list_very_special(n)
   z = [u[k] ** 2 - 2 for k in range(n)]
   lhs = sz_via_jt1(z, [m] * n)
   powers1 = [m + 2 * j + 1 for j in range(n)]
   powers2 = [m + 2 * j for j in range(n)]
   powers3 = [2 * j + 1 for j in range(n)]
   powers4 = [2 * j for j in range(n)]
   numer = tdet(u, powers1) * udet(u, powers2)
   denom = tdet(u, powers3) * udet(u, powers4)
   rhs = numer / denom
   if verbos >= 3:
      print('lhs = ' + str(lhs))
      print('rhs = ' + str(rhs))
   return lhs == rhs

def big_test_trench_numer(nmax, mmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_trench_numer, nmax = %d, mmax = %d' % (nmax, mmax))
   tasks = [(n, m, r) for n in range(1, nmax + 1) for m in range(mmax + 1) for r in range(nrep)]
   return all(test_trench_numer(n, m, verbos) for (n, m, r) in tasks)

import time

def all_big_tests_trench_krattenthaler_elouafi_numer():
   nrep = 10
   verbos = 2
   t = time.time()
   res = [0] * 3
   res[0] = big_test_krattenthaler_numer(7, 20, nrep, verbos)
   res[1] = big_test_elouafi_numer(8, 20, nrep, verbos)
   res[2] = big_test_trench_numer(10, 25, nrep, verbos)
   total_result = all(res)
   t = time.time() - t
   print('time = %.2f seconds' % t)
   print('total result = ' + str(total_result))

# time < 200 sec
all_big_tests_trench_krattenthaler_elouafi_numer()

